local noise = require("noise")
local tne = noise.to_noise_expression
local litexp = noise.literal_expression

local seed1 = 1234
local pi_over_3 = tne(math.pi)/3


-- distance of point from center in radius units
local function dist(x0, y0, r)
  local center = noise.make_point_list({{x0, y0}})
  return noise.distance_from(noise.var("x"), noise.var("y"), center) / r
end

-- circle fades till radius r
local function circle_expression(x0, y0, r)
  local center = noise.make_point_list({{x0, y0}})
  local dist = noise.distance_from(noise.var("x"), noise.var("y"), center)
  return noise.max(0, 1 - dist / r)
end

-- circle with radius r1, fades till radius r2
local function circle_expression2(x0, y0, r1, r2)
  local center = noise.make_point_list({{x0, y0}})
  local dist = noise.distance_from(noise.var("x"), noise.var("y"), center)
  return noise.max(0, 1 - noise.max(0, dist - r1) / (r2 - r1))
end

--[[
factorio-multioctave-noise

Arguments (named):

    x
    y
    seed0 - integer between 0 and 4294967295 (inclusive) used to populate the backing random noise
    seed1 - integer between 0 and 255 (inclusive) used to provide extra randomness when sampling
    octaves - how many layers of noise at different scales to sum
    persistence (constant) - how strong is each layer compared to the next larger one
    input_scale (default: 1) - x and y will be multiplied by this before sampling
    output_scale (default: 1) - output will be multiplied by this before being returned
]]--

local seed = 1234
local scale = 4
local land = tne{
  type = "function-application",
  function_name = "factorio-multioctave-noise",
  arguments = {
    x = noise.var("x"),
    y = noise.var("y"),
    seed0 = noise.var("map_seed"), -- i.e. map.seed
    seed1 = tne(seed), -- Some random number
    octaves = tne(6),
    persistence = tne(0.5),
    input_scale = noise.var("segmentation_multiplier") / scale,
    output_scale = scale / noise.var("segmentation_multiplier"),
  }
}

local spots = function()
  local seed1 = 123
  local pi_over_3 = tne(math.pi)/3

  -- enemy_base_border_width gives the width of the border in which probability ramps up from 0 to 1.
  -- i.e. slope = 1 / border_width, and height = radius / border_width
  -- In practice, a probabilty of about 1/9 will result in a solid carpet of biter bases,
  -- which is why this border width is so high (ignoring blobbiness, most of the inner part of the 'border' would be solid biter bases)
  local enemy_base_border_width = 30

  -- Spots per square meter (should be a very small fraction):
  local frequency_expression = noise.max(0, noise.var("enemy-base-frequency"))
  local spot_radius_expression = noise.max(0, noise.var("enemy-base-radius"))
  -- multiply by spot_radius_expression to keep slope constant:
  local spot_height_expression = spot_radius_expression / enemy_base_border_width
  local spot_quantity_expression = pi_over_3 * spot_radius_expression * spot_radius_expression * spot_height_expression
  -- spot_quantity_expression = noise.compile_time_log("Spot quantity: ", spot_quantity_expression)
  local density_expression = spot_quantity_expression * frequency_expression
  -- density_expression = noise.compile_time_log("Density: ", density_expression)

  local basement_value = -4

  local spots = tne{
    type = "function-application",
    function_name = "spot-noise",
    arguments = {
      x = noise.var("x"),
      y = noise.var("y"),
      seed0 = noise.var("map_seed"),
      seed1 = tne(seed1),
      region_size = tne(512),
      candidate_point_count = tne(100),
      density_expression = litexp(density_expression), -- low-frequency noise evaluate for an entire region
      spot_quantity_expression = litexp(spot_quantity_expression), -- used to figure out where spots go
      hard_region_target_quantity = tne(false), -- it's fine for large spots to push region quantity past the target
      spot_radius_expression = litexp(spot_radius_expression),
      spot_favorability_expression = litexp(1),
      basement_value = tne(basement_value),
      maximum_spot_basement_radius = tne(128)
    }
  }
  return spots -- + circle_expression(0, 0, 128)
end

--[[
local island = (land + 3) * circle_expression2(0, 0, 128, 512) - 3
data:extend{
  {
    type = "noise-expression",
    name = "CS-islands",
    intended_property = "elevation",
    expression = noise.define_noise_function(function(x, y, tile, map)
        return noise.if_else_chain(
          island, 1,
          noise.min(1, dist(0, 0, 512))) * island
    end)
  }
}
]]--

-- normalize spots to have a circle of 1 inside and fade to 0
local spots_land_factor = noise.min(4, spots() + 4) / 4

-- fade out islands as they leave spots
local island = (land + 3) * spots_land_factor - 3

-- normalize spots to have 0 at center and rise to 5/3
local spots_water_factor = (1 - spots()) / 3

data:extend{
  {
    type = "noise-expression",
    name = "CS-islands",
    intended_property = "elevation",
    expression = noise.define_noise_function(function(x, y, tile, map)
        -- factor in water factor around islands
        return noise.if_else_chain(
          island, 1,
          spots_water_factor) * island
    end)
  }
}

data.raw["map-gen-presets"]["default"].islands = {
  order = "i",
  basic_settings = {
    property_expression_names = {
      elevation = "CS-islands",
    },
    autoplace_controls = {},
    terrain_segmentation = 1,
  }
}
